import numpy as np

class OU:
    """ Implements one-dimensional Ornstein-Uhlenbeck process.
     The potential is given by

        V(x) = 0.5* K * (x - c)**2

    """

    def __init__(self, K=1.0, c=0.0, beta=1.0):
        """

        Parameters:
        ------------
        K, float:       Pre-factor to set stiffness of the potential
        c, float:       location of the minimum
        beta, float:    inverse temperature.
        """
        # Get info:
        self.K = K
        self.c = c
        self.beta = beta

    def potential(self, x):
        """
        Evaluate potential energy at positions x

        x, nd-array (m,) or (1, m):
            Array of coordinates.

        Returns:
        --------
        V, nd-array (m):
            Values of the potential for x-values.
        """

        return 0.5 * self.K * (x - self.c) ** 2

    def gradient(self, x):
        """
        Evaluate gradient of energy at positions x

        x, nd-array (m,) or (1, m):
            Array of coordinates.

        Returns:
        --------
        dV, nd-array (1, m):
            Values of the potential gradient for positions in x.
        """
        # Get info:
        if x.ndim == 1:
            x = x[None, :]
        m = x.shape[1]
        # Compute derivatives:
        dV = self.K * (x - self.c)

        return dV

    def simulate(self, x0, m, dt):
        """
            Generate trajectory of the overdamped Langevin dynamics, using first order Euler scheme.

            x0, nd-array(d,) or (msim, d):
                initial values for the simulation. This can be a single d-dimensional vector, or an
                array of msim initial position in d-dimensional space. In the latter case, msim
                independent simulations will be produced.
            m, int:
                number of time steps to be returned (not including initial value)
            dt, float:
                integration time step

            Returns:
            --------
            X, nd-array (d, m+1) or (msim, d, m+1):
                simulation trajectory(ies)
            """
        # Determine shape of initial data:
        if len(x0.shape) == 1:
            msim = 1
            x0 = x0[None, :]
        else:
            msim = x0.shape[0]
        d = x0.shape[1]
        # Initialize:
        X = np.zeros((msim, d, m + 1))
        # Fill in initial values:
        X[:, :, 0] = x0.copy()
        X_old = x0.T

        # Run simulation:
        for t in range(1, m + 1):
            # Euler-Maruyama update step:
            X_new = X_old - self.gradient(X_old) * dt + \
                    np.sqrt(2 * dt / self.beta) * np.random.randn(d, msim)
            # Store next position:
            X[:, :, t] = (X_new.T).copy()
            # Overwrite:
            X_old = X_new

        # Trim shape of X if only one simulation was requested:
        if msim == 1:
            X = X[0, :, :]

        return X

    def stat_dist(self, x):
        """
        Evaluate stationary density at positions x

        x, nd-array (m,) or (1, m):
            Array of coordinates.

        Returns:
        --------
        mu, nd-array (1, m):
            Values of the Boltzmann distribution for positions in x.
        """
        # Get info:
        if x.ndim == 1:
            x = x[None, :]
        m = x.shape[1]
        # Return stationary density:
        return np.sqrt((self.K*self.beta) / (2*np.pi)) * np.exp(-0.5*self.K * self.beta * (x - self.c)**2)
