import numpy as np
import matplotlib.pyplot as plt
from scipy.special import eval_hermite, gamma, factorial
from scipy.integrate import dblquad

import json

import OU
from mercer import Mercer

""" ------- SETTINGS ----------"""
# Load parameters from .json file:
json_file = open("parameters.json", "r")
param = json.load(json_file)

""" Settings related to OU process:"""
# Set spring constant and temperature for OU Process:
alpha = param["alpha"]
beta = param["beta"]

# Create instance of abstract OU class:
ou = OU.OU(alpha, beta=beta)

""" Settings related to Koopman estimation: """
# Elementary time step for simulations:
dt = param["dt"]
# Estimation time for Koopman operator:
tint = param["tint"]
t = dt * tint
# Number of spectral components of the Koopman operator to be used in error bounds:
Jmax = param["Jmax"]

""" Settings related to the kernel: """
# List of kernel bandwidths:
sigma_list = param["sigma_list"]
# Number of Mercer features to be used in calculation of error bounds:
nmax_list = param["nmax"]

""" Experimental Settings: """
# Upper and lower bound for numerical integrals,
# needed for the calculation of error bounds:
Lint = param["Lint"]

""" ---------- Basic Functions and quantities: ------------- """
# Function to evaluate condional transition density (analytically known for OU process):
def _rho_t(x, y, t):
    nu_t = 1.0 - np.exp(-2*alpha*t)
    return np.sqrt(alpha * beta/(2*np.pi * nu_t)) * np.exp(-0.5*alpha*beta / nu_t * (y - np.exp(-alpha*t)*x)**2)

# Joint probability density for two observations, separated by time t:
def _nu_t(x, y, t):
    return ou.stat_dist(np.array([x]))[0, :] * _rho_t(x, y, t)

# Compute eigenvalues and eigenfunctions of the Koopman operator at time t according to analytical formula:
q = np.array([np.exp(-alpha * jj * dt) for jj in range(Jmax)])
psi = lambda x, l: np.sqrt(1.0/(2**l * alpha**l *factorial(l))) * eval_hermite(l, np.sqrt(2*alpha)*x)

""" ----------------- Computation of fundamental quantities: ------------------"""
# Iterate over all bandwidths:
qq = 0
for sigma in sigma_list:
    print("Processing bandwidth sigma = %.2f..."%sigma)
    # Number of Mercer features for this bandwidth:
    nmax = nmax_list[qq]
    # Define class holding Mercer features:
    merc = Mercer(sigma, alpha)
    # Functions for eigenvalues and eigenfunctions:
    phi = merc.phi
    lambda_n = merc.lambda_n

    """ Integrals for functions g_{ij}"""
    print("Computing integrals for functions g_ij...")
    MJ = np.zeros((nmax, nmax, Jmax))
    for ii in range(nmax):
        for jj in range(nmax):
            for rr in range(Jmax):
                MJ[ii, jj, rr] = dblquad(lambda x, y: phi(x, ii) * phi(y, jj) * psi(x, rr) * _nu_t(x, y, t),
                                         -Lint, Lint, -Lint, Lint)[0]
        print("Completed ii = %d" % (ii + 1))

    """ Calculaton of d_{r,t}:"""
    drt = np.zeros(Jmax)
    for rr in range(Jmax):
        drt[rr] = np.dot(lambda_n(range(nmax)).T, np.dot(MJ[:, :, rr] * MJ[:, :, rr].T, lambda_n(range(nmax))))
    print("Resulting d-vector:", drt)

    """ Hilbert-Schmidt norm of C^t_H: """
    ct_norm = np.sqrt(np.dot(lambda_n(range(nmax)).T, np.dot(MJ[:, :, 0] ** 2, lambda_n(range(nmax)))))
    print("Hilbert-Schmidt norm for C^t_H: ", ct_norm)

    """ Save analytical quantities to disc: """
    print("Saving results...")
    di = {}
    di["ct_norm"] = ct_norm
    di["MJ"] = MJ
    di["drt"] = drt
    np.savez("Data/AnalyticalQuantities_t=%.2f_sigma=%.3f.npz"%(t, sigma), **di)
    print("Done.")

    """ ------------- VISUALIZATION ----------------"""
    # One-dimensional grid for plots:
    xp = np.arange(-4.0, 4.01, 0.02)
    """ Visualize Mercer Eigenvalues and Features: """
    plt.figure(figsize=(12, 8))
    plt.subplot(2, 2, 1)
    for ll in range(nmax):
        plt.plot(xp, phi(xp, ll), "--")
    plt.ylim([-10.0, 10.0])
    plt.title(r"Mercer Features $\sigma = %.2f$"%sigma)

    plt.subplot(2, 2, 2)
    plt.plot(range(nmax), lambda_n(range(nmax)), "o--")
    plt.yscale("log")
    plt.ylim([1e-3, 1e0])
    plt.title("Mercer Eigenvalues $\sigma = %.2f$"%sigma)

    """ Visualize Koopman Eigenpairs: """
    plt.subplot(2, 2, 3)
    for ll in range(0, Jmax, 2):
        plt.plot(xp, psi(xp, ll) * ou.stat_dist(xp)[0, :], "--")
    plt.title("Koopman Eigenfunctions t=%.2f"%t)

    plt.subplot(2, 2, 4)
    plt.plot(range(Jmax), q, "ro--")
    plt.ylim([0.5, 1.05])
    plt.title(r"Koopman Eigenvalues at $t = %.2f$"%t)

    qq += 1