import numpy as np
from scipy.integrate import quad
from scipy.linalg import eigh

import OU
from mercer import Mercer
from gaussian_kernel import gaussian_kernel

import time
import json


""" In this script, we compute estimation errors and prediction errors for different data sizes
    and kernel bandwidths. See README for details.
"""

""" Function to compute estimation error for given independent trajectories of fixed length."""
def _estimation_error(Xf, phi, lambda_n, M, ct_n):
    # Obtain shape of input data:
    ntraj, m = Xf.shape
    # Obtain number of Mercer features:
    nmax = M.shape[0]
    # Reduce data size by one step:
    m = m - 1

    # Compute estimation error for each trajectory:
    E = np.zeros(ntraj)
    for kk in range(ntraj):
        # Extract time-lagged slices of the data:
        X = Xf[kk, :-1]
        Y = Xf[kk, 1:]

        # Evaluate Mercer features on data points:
        Phi_X = np.vstack([phi(X, ll) for ll in range(nmax)])
        Phi_Y = np.vstack([phi(Y, ll) for ll in range(nmax)])

        # Compute empirical approximation of M-matrix:
        Mhat = (1.0 / m) * np.dot(Phi_X, Phi_Y.T)

        # HS-norm of empirical co-variance operator:
        ct_hat_n = np.sqrt(np.dot(lambda_n(range(nmax)).T, np.dot(Mhat ** 2, lambda_n(range(nmax)))))
        # Cross-terms between empirical and analytical co-variance ops
        ct_cross = np.dot(lambda_n(range(nmax)).T, np.dot(Mhat * M, lambda_n(range(nmax))))

        # Compute estimation error:
        E[kk] = np.sqrt(ct_n ** 2 + ct_hat_n ** 2 - 2 * ct_cross)
    return E

""" Function to compute prediction error from simulation data: """
def _prediction_error(Xf, f0, kg, Ktf0_N, nmax):
    # Obtain shape of the data:
    ntraj_p, m = Xf.shape
    m = m - 1
    # Output:
    Ep = np.zeros(ntraj_p)
    # Iterate over trajectories:
    for rr in range(ntraj_p):
        # Split data into X and Y parts:
        Xkk = Xf[rr, :-1][None, :]
        Ykk = Xf[rr, 1:][None, :]
        # Evaluate and diagonalize kernel matrix for each trajectory:
        KXX = kg(Xkk, Xkk)
        lx, Vx = eigh((1.0/m) * KXX, subset_by_index=[m-nmax, m-1])
        lx = lx[::-1]
        Vx = Vx[:, ::-1]
        # Inverted projection matrix:
        Vp = Vx @ ((1.0/lx)[:, None] * Vx.T)
        # Define callable for truncated empirical prediction of f0:
        Ktphi_hat = (1.0/m) * np.dot(f0(Ykk), Vp)
        Ktf0_hat = lambda x: np.dot(Ktphi_hat, kg(Xkk, np.array([x])[None, :]))
        # Compute error:
        Ep[rr] = np.sqrt(quad(lambda x: (Ktf0_N(x) - Ktf0_hat(x))**2 * ou.stat_dist(np.array([x]))[0, :],
                                  -5.0, 5.0)[0])

    return Ep

""" ------- SETTINGS ----------"""
# Load parameters from .json file:
json_file = open("parameters.json", "r")
param = json.load(json_file)
""" Settings related to OU process:"""
# Set spring constant and temperature for OU Process:
alpha = param["alpha"]
beta = param["beta"]

# Create instance of abstract OU class:
ou = OU.OU(alpha, beta=beta)

""" Settings related to Koopman estimation: """
# Elementary time step for simulations:
dt = param["dt"]
# Estimation time for Koopman operator:
tint = param["tint"]
t = dt * tint

""" Settings related to the kernel: """
# List of kernel bandwidths:
sigma_list = param["sigma_list"]
# Number of Mercer features to be used in calculation of error bounds:
nmax_list = param["nmax"]

""" Experimental Settings: """
# Array of data sizes for estimation error:
marr = np.array(param["marr"])
num_m = marr.shape[0]
m_max = marr[-1]
# Confidence level:
delta = 1 - param["conf"]

# Number of trajectories for empirical error estimation:
ntraj = param["ntraj"]
# Set data sizes used for prediction:
marr_p = np.array(param["marr_p"])
num_mp = marr_p.shape[0]
# Number of independent trajectories for prediction:
ntraj_p = param["ntraj_p"]



""" ---------- COMPUTE DATA-BASED ERROR ESTIMATES ------------- """
qq = 0
for sigma in sigma_list:
    print("Processing sigma = %.2f..."%sigma)
    # Number of Mercer features for this bandwidth:
    nmax = nmax_list[qq]
    # Creat instance of kernel object:
    kg = gaussian_kernel(sigma)
    """ Preparations: """
    # Load analytically computed quantities:
    di = np.load("Data/AnalyticalQuantities_t=%.2f_sigma=%.3f.npz"%(t, sigma))
    M = di["MJ"][:, :, 0]
    ct_n = di["ct_norm"]

    # Create Mercer object:
    merc = Mercer(sigma, alpha)

    """ Sampling """
    # Draw intial conditions from invariant distribution:
    y0_ = np.sqrt(1.0/(alpha*beta)) * np.random.randn(ntraj, 1)
    # Generate simulation data:
    Xfull = ou.simulate(y0_, m_max*tint, dt)[:, 0, ::tint]
    print("Produced simulation data of shape: ", Xfull.shape)

    # Compute estimation error for all trajectories and different data sizes:
    E_list = np.zeros((ntraj, num_m))
    print("Computing estimation errors...")
    for ii in range(num_m):
        m = marr[ii]
        E_list[:, ii] = _estimation_error(Xfull[:, :m+1], merc.phi, merc.lambda_n, M, ct_n)
        print("Completed m = %d"%m)

    """ Prepare analytical expression for application of Koopman operator to a fixed
        observable.
     """
    print("Computing prediction errors...")
    # Define Gaussian observable, equal to the first Mercer feature (normalized as probability density)
    mu0 = 0.0
    sig_0 = 1.0/(np.sqrt(2)*merc.zeta)
    f0 = lambda x: (1.0/np.sqrt(2*np.pi*sig_0**2)) * np.exp(-0.5*((x - mu0)/sig_0)**2)

    # Callable function for application of Koopman operator to f0:
    nu_t = np.sqrt((1.0 - np.exp(-2*alpha*t))/(alpha*beta))
    sig_mix = np.sqrt(nu_t**2 + sig_0**2)
    Ktf0 = lambda x: ((1.0/np.sqrt(2*np.pi * sig_mix**2)) *
                               np.exp(-0.5*((mu0 - np.exp(-alpha*t) * x)/sig_mix)**2))
    # Compute its coefficients w.r.t. Mercer features:
    ct = np.zeros(nmax)
    for jj in range(nmax):
        ct[jj] = quad(lambda x: Ktf0(x) * merc.phi(x, jj) * ou.stat_dist(np.array([x]))[0, :],
                      -5.0, 5.0)[0]
    # Callable for truncated solution:
    Ktf0_N = lambda x: np.sum(ct * np.array([merc.phi(x, ll) for ll in range(nmax)]))

    """ Numerical estimation of the prediction error for f0: """
    # Generate independent simulation data:
    y0_ = np.sqrt(1.0 / (alpha * beta)) * np.random.randn(ntraj_p, 1)
    Xfull = ou.simulate(y0_, marr_p[-1] * tint, dt)[:, 0, ::tint]
    print("Produced simulation data of shape ", Xfull.shape)
    # Prepare output matrices:
    Ep = np.zeros((num_mp, ntraj_p))
    # Compute prediction error for each data size and all trajectories:
    for kk in range(num_mp):
        m = marr_p[kk]
        t1 = time.time()
        Ep[kk, :] = _prediction_error(Xfull[:, :m+1], f0, kg, Ktf0_N, nmax)
        t2 = time.time()
        print("Completed m = %d after %.2f secs"%(m, t2-t1))

    qq += 1

    """ Save to disc: """
    print("Saving results...")
    di = {}
    di["E_est"] = E_list
    di["E_predict"] = Ep
    np.savez("Data/Empirical_Errors_t=%.2f_sigma=%.3f.npz"%(t, sigma), **di)