import numpy as np
import matplotlib.pyplot as plt

import json

""" This script generates the figures shown in the main paper."""

"""-------- SETTINGS: -----"""
# Load parameters from .json file:
json_file = open("parameters.json", "r")
param = json.load(json_file)
""" Settings related to Koopman estimation: """
# Elementary time step for simulations:
dt = param["dt"]
# Estimation time for Koopman operator:
tint = param["tint"]
t = dt * tint

""" Experimental Settings: """
# Array of data sizes for estimation error:
marr = np.array(param["marr"])
num_m = marr.shape[0]
m_max = marr[-1]
# Confidence level:
delta = 1 - param["conf"]
# Set data sizes used for prediction:
marr_p = np.array(param["marr_p"])
num_mp = marr_p.shape[0]

# List of kernel bandwidths to display:
sigma_list = param["sigma_list"]

""" -------- GENERATE PLOTS -----------"""

markers = ["o", "x", "s"]
f1 = plt.figure(figsize=(6, 4))
f2 = plt.figure(figsize=(6, 4))
f3 = plt.figure(figsize=(6, 4))

plt.rcParams['font.size'] = 16
plt.rc('legend', fontsize=12)

qq = 0
for sigma in sigma_list:
    # Load theoretical results:
    di = np.load("Data/Error_Bounds_t=%.2f_sigma=%.3f.npz" % (t, sigma))
    E_iid = di["E_iid"]
    E_erg = di["E_erg"]
    E_cor = di["E_cor"]
    E_pr = di["E_pr"]
    # Load empirical results:
    di_e = np.load("Data/Empirical_Errors_t=%.2f_sigma=%.3f.npz"%(t, sigma))
    E_est = di_e["E_est"]
    E_predict = di_e["E_predict"]

    """ Show results for estimation error:"""
    plt.figure(f1)
    # Plot ergodic bounds vs data error:
    plt.plot(marr, E_erg, "b%s--"%markers[qq], linewidth=2.5, 
             label=r"$\sigma = %.2f$"%sigma)
    plt.plot(marr, np.percentile(E_est, 100 * (1 - delta), axis=0), "r%s--"%markers[qq],
             linewidth=2.5)

    plt.figure(f2)
    # Plot iid and corollary bounds vs data error:
    plt.plot(marr, E_cor, "g%s--"%markers[qq], linewidth=2.5, label=r"$\sigma = %.2f$"%sigma)
    plt.plot(marr, E_iid, "m%s--"%markers[qq], linewidth=2.5)
    plt.plot(marr, np.percentile(E_est, 100 * (1 - delta), axis=0), "r%s--"%markers[qq],
             linewidth=2.5)

    """ Show results for prediction error:"""
    plt.figure(f3)
    # Theoretical Bound:
    plt.errorbar(marr_p, np.mean(E_predict, axis=1), yerr=np.std(E_predict, axis=1), 
                 fmt="r%s--"%markers[qq], linewidth=2.5)
    plt.plot(marr_p, E_pr[:num_mp], "b%s--"%markers[qq], linewidth=2,
              label=r"$\sigma = %.2f$"%sigma)
    
    qq += 1

# Format first figure:
plt.figure(f1)
ax = plt.gca()
plt.xlabel(r"$m$")
ax.xaxis.set_label_coords(0.65, -0.05)
plt.ylabel(r"$\varepsilon$")
plt.xscale("log")
plt.yscale("log")
ax.yaxis.set_label_coords(-.05, 0.75)
plt.ylim([1e-3, 5e1])
plt.title(r"Thm 3.1 vs Data")
plt.legend(loc=3, frameon=True)
plt.savefig("Figures/Ergodic_v_Data.pdf", bbox_inches="tight")

# Format second figure
plt.figure(f2)
ax = plt.gca()
plt.xlabel(r"$m$")
ax.xaxis.set_label_coords(0.65, -0.05)
plt.ylabel(r"$\varepsilon$")
plt.xscale("log")
plt.yscale("log")
ax.yaxis.set_label_coords(-.05, 0.75)
plt.ylim([1e-3, 5e1])
plt.title(r"I.I.D. and Rem. 3.4 vs Data")
plt.legend(loc=3, frameon=True)
plt.savefig("Figures/IID_vs_Data.pdf", bbox_inches="tight")

# Format third figure:
plt.figure(f3)
ax = plt.gca()
plt.xlabel(r"$m$")
ax.xaxis.set_label_coords(0.5, -0.05)
plt.ylabel(r"$\varepsilon$")
plt.xscale("log")
plt.yscale("log")
ax.yaxis.set_label_coords(-.05, 0.60)
plt.ylim([5e-3, 1e8])
plt.title(r"Prediction Error")
plt.legend(loc=1, frameon=True)
plt.savefig("Figures/Prediction_Error.pdf", bbox_inches="tight")


plt.show()