import numpy as np
from scipy.spatial.distance import pdist, cdist, squareform

class gaussian_kernel:

    def __init__(self, sigma):
        """
        Implements standard Gaussian RBF kernel in d-dimensional space.

        Parameters:
        ----------
        sigma, float:       kernel bandwidth.
        """
        self.sigma = sigma

    def __call__(self, X, Y=None):
        """
        Evaluate the kernel for all pairs in given data set(s).

        Parameters:
        ----------
        X, (d, mx)           data array, m data points in d-dimensional space
        Y, (d, my)           second data array, m data points in d-dimensional space.
                             If Y is not None, evaluate k(x_i, y_j) for all (i, j),
                             otherwise k(x_i, x_j) for all (i, j) is returned.

        Returns:
        --------
        KX, (mx, mx) or (mx, my):
                            kernel Gramian (evaluation at all pairs of data sites)
        """
        # Compute matrix of squared pair-wise Euclidean distances using library function:
        if Y is None:
            DX = squareform(pdist(X.T, metric="sqeuclidean"))
        else:
            DX = cdist(X.T, Y.T, metric="sqeuclidean")

        return np.exp(-(1.0 / self.sigma**2) * DX)