import numpy as np
import json

from mercer import Mercer

""" In this script, we calculate theoretical error bounds we derived in the paper.
    We use three different kernel bandwidths and a range of data sizes $m$.
"""


""" ------- SETTINGS ----------"""
# Load parameters from .json file:
json_file = open("parameters.json", "r")
param = json.load(json_file)

""" Settings related to OU process:"""
# Set spring constant and temperature for OU Process:
alpha = param["alpha"]
""" Settings related to Koopman estimation: """
# Elementary time step for simulations:
dt = param["dt"]
# Estimation time for Koopman operator:
tint = param["tint"]
t = dt * tint
# Number of Koopman eigenvalues to use:
Jmax = param["Jmax"]

""" Settings related to the kernel: """
# List of kernel bandwidths:
sigma_list = param["sigma_list"]
# Number of Mercer features to be used in calculation of error bounds:
nmax_list = param["nmax"]
""" Experimental Settings: """
# Array of data sizes:
marr = np.array(param["marr"])
num_m = marr.shape[0]
# Confidence level:
delta = 1 - param["conf"]

""" ---------- CALCULATION OF ERROR BOUNDS --------------"""
# Analytical eigenvalues of the Koopman operator at chosen lag time:
q = np.array([np.exp(-alpha * jj * t) for jj in range(Jmax)])
# Function applied to eigenvalues:
def _Fm(z, m):
    return (2.0/(1 - z)) * (1.0 - (1.0 - z**m)/(m * (1.0 - z)))

# Iterate over all kernel bandwidths:
qq = 0
for sigma in sigma_list:
    # Number of Mercer features for this bandwidth:
    nmax = nmax_list[qq]
    # Load analytically computed quantities:
    di = np.load("Data/AnalyticalQuantities_t=%.2f_sigma=%.3f.npz"%(t, sigma))
    MJ = di["MJ"]
    drt = di["drt"]
    ct_norm = di["ct_norm"]

    # Define class holding Mercer features:
    merc = Mercer(sigma, alpha)

    # Spectral gap constant for kernel integral operator:
    delta_nmax = np.min(0.5*(merc.lambda_n(range(nmax+1))[:-1] - merc.lambda_n(range(nmax+1))[1:]))

    """ Calculate theoretical error bounds based on three different variances: """
    # Output matrices
    E_iid = np.zeros((num_m,))
    E_erg = np.zeros((num_m,))
    E_cor = np.zeros((num_m,))
    E_pr = np.zeros((num_m,))

    # Solve Chebyshev inequality for error using all three variances:
    for i in range(num_m):
        m = marr[i]
        # iid
        sigma_i = 1.0 - ct_norm ** 2
        E_iid[i] = np.sqrt(sigma_i / (delta * m))

        # ergodic variance
        Fmq = _Fm(q, m)
        sigma_e = (1.0 - ct_norm ** 2) + np.dot(drt[1:], Fmq[1:])
        E_erg[i] = np.sqrt(sigma_e / (delta * m))

        # Corollary 3.3:
        sigma_c = (1.0 - ct_norm ** 2) * (3 - q[1])/(1 - q[1])
        E_cor[i] = np.sqrt(sigma_c / (delta * m))

        # Prediction error according to Theorem 4.1:
        E_pr[i] = ((1.0 / np.sqrt(merc.lambda_n(nmax))) + 2 * (nmax + 1) / (delta_nmax * merc.lambda_n(nmax))) * E_erg[i]

    qq += 1

    # Save results:
    di = {}
    di["E_iid"] = E_iid
    di["E_erg"] = E_erg
    di["E_cor"] = E_cor
    di["E_pr"] = E_pr
    np.savez("Data/Error_Bounds_t=%.2f_sigma=%.3f.npz"%(t, sigma), **di)